#!/usr/bin/php
<?php
// A virtual meters example for meterN. This script will simulate a house and self consumuption estimated meters (*). You must own a total (import/export) and a production meter. 
//                         _____  
//                        /     \
//      +----------+     /       \                                  - ^ -
//      |Production| --> | House | <-- import ___ +-----------+ ___  /X\ Grid
//      +----------+     |_______| --> export     |Total meter|     /V V\
//                                                +-----------+
//                consumption/selfconsumuption (*)
//
// ln -s /srv/http/comapps/houseenergy.php /usr/bin/houseenergy
// houseenergy [ power | powerimp | powerexp | powerself | energy | self ]
//
// The house virtual meter should be configured in mN as 'Elect House consumption' type with a passover value like 100000.
// The house self consumuption meter as 'Elect Other' with a passover value like 100000.

if (isset($_SERVER['REMOTE_ADDR'])) {
    die('Direct access not permitted');
}
// Set the virtual meters
// Consumption
$HOUSEID     = 'elect'; // ID
$HOUSEmetnum = 1; // meter number
// Selfconsumption
$SELFID      = 'self';
$SELFCmetnum = 7;

// Set up real meters
// Production
$PRODID       = 'solar'; // ID
$PRODcmd      = 'pool123s energy'; // Energy command
$POWERPRODcmd = 'pool123s power'; // Power
$PRODmetnum   = 4; // meter number
// TOT return the total power (eg if import = 45W, export -55W)
$TPID         = '1_P';
$TOTPOWERcmd  = 'cat /dev/shm/metern_log.txt | egrep "^1_P\(" | grep "*W)"';
//$TOTPOWERcmd  = 'sdm120c -a1 -d2 -b9600 -2 -z3 -m -p /dev/sdm';
// Energy Imported
$IMPID        = '1_IE';
$IMPcmd       = 'cat /dev/shm/metern_log.txt | egrep "^1_IE\(" | grep "*Wh)"';
//$IMPcmd       = 'sdm120c -a1 -d2 -b9600 -2 -z3 -m -i /dev/sdm';
$IMPmetnum    = 5;
// Energy Exported
$EXPID        = '1_EE';
$EXPcmd       = 'cat /dev/shm/metern_log.txt | egrep "^1_EE\(" | grep "*Wh)"';
//$EXPcmd       = 'sdm120c -a1 -d2 -b9600 -2 -z3 -m -e /dev/sdm';
$EXPmetnum    = 6;

// Path to metern
$MNDIR = '/srv/http/metern';

// No edit should be needed bellow
$prevfile = '/dev/shm/prevhouseenergy.json';

function getvalue($id, $cmd) //  Get data and validate with IEC 62056 data set structure
{
    $datareturn = null;
    $giveup     = 0;
    $regexp     = "/^$id\(-?[0-9\.]+\*[A-z0-9%]+\)$/i"; //ID(VALUE*UNIT)
    
    while (!isset($datareturn) && $giveup < 3) { // Try 3 times
        exec($cmd, $datareturn);
        $datareturn = trim(implode($datareturn));
        
        if (preg_match($regexp, $datareturn)) {
            $datareturn = preg_replace("/^$id\(/i", '', $datareturn, 1); // VALUE*UNIT)
            $datareturn = preg_replace("/\*[A-z0-9%]+\)$/i", '', $datareturn, 1); // VALUE
            settype($datareturn, 'int');
        } else {
            $datareturn = null;
        }
        $giveup++;
    }
    return $datareturn;
}

function retrievecsv($meternum, $csvarray, $passo, $datareturn) // Retrieve last know value in latest csv
{
    $datareturn = null;
    $contalines = count($csvarray);
    $j          = 0;
    while (!isset($datareturn)) {
        $j++;
        $array      = preg_split('/,/', $csvarray[$contalines - $j]);
        $datareturn = (int) trim($array[$meternum]);
        if ($datareturn == '') {
            $datareturn = null;
        }
        if ($j == $contalines) {
            $datareturn = 0;
        }
    }
    if ($datareturn > $passo) {
        $datareturn -= $passo;
    }
    return $datareturn;
}

if (isset($argv[1])) {
    if ($argv[1] == 'power' || $argv[1] == 'powerimp' || $argv[1] == 'powerexp' || $argv[1] == 'powerself') { // Power
        $totpower  = getvalue($TPID, $TOTPOWERcmd);
        $prodpower = getvalue($PRODID, $POWERPRODcmd);
        $power     = $prodpower + $totpower;
        
        if ($argv[1] == 'power') {
            $outstr = utf8_decode("$HOUSEID($power*W)\n");
            //usleep(500000);
        } else {
            if ($totpower > 0) { // Import
                $imppower = $power - $prodpower;
                $exppower = 0;
                $slfpower = $prodpower;
            } else { // Export
                $imppower = 0;
                $exppower = $prodpower - $power;
                $slfpower = $power;
            }
        }
        if ($argv[1] == 'powerimp') {
            $outstr = utf8_decode("$IMPID($imppower*W)\n");
        } elseif ($argv[1] == 'powerexp') {
            $outstr = utf8_decode("$EXPID($exppower*W)\n");
        } elseif ($argv[1] == 'powerself') {
            $outstr = utf8_decode("$SELFID($slfpower*W)\n");
        }
        echo "$outstr";
    } elseif ($argv[1] == 'energy' || $argv[1] == 'self') { // Energy
        define('checkaccess', TRUE);
        include("$MNDIR/config/config_main.php");
        include("$MNDIR/config/config_met$HOUSEmetnum.php");
        include("$MNDIR/config/config_met$SELFCmetnum.php");
        include("$MNDIR/config/config_met$IMPmetnum.php");
        include("$MNDIR/config/config_met$EXPmetnum.php");
        include("$MNDIR/config/config_met$PRODmetnum.php");
        
        // Retrieve previous virtuals meters values
        if (file_exists($prevfile)) {
            $data     = file_get_contents($prevfile);
            $previous = json_decode($data, true);
        } else {
            $tdfile  = date('Ymd');
            $tdfile  = $MNDIR . '/data/csv/' . $tdfile . '.csv';
            $ytdfile = date('Ymd', strtotime("-1 day"));
            $ytdfile = $MNDIR . '/data/csv/' . $ytdfile . '.csv';

            if (file_exists($tdfile)) {
                $lines = file($tdfile);
            } else if (file_exists($ytdfile)) {
                $lines = file($ytdfile);
            }
            
            if (file_exists($tdfile) || file_exists($ytdfile)) { // today or yesterday
                $contalines = count($lines);
                $lastarray  = preg_split("/,/", $lines[$contalines - 1]);
                
                $datareturn                = null;
                $previous['prevIMPhouse']  = retrievecsv($IMPmetnum, $lines, ${'PASSO' . $IMPmetnum}, $datareturn);
                $datareturn                = null;
                $previous['prevEXPhouse']  = retrievecsv($EXPmetnum, $lines, ${'PASSO' . $EXPmetnum}, $datareturn);
                $previous['prevEXPself']   = $previous['prevEXPhouse'];
                $datareturn                = null;
                $previous['prevHOUSE']     = retrievecsv($HOUSEmetnum, $lines, ${'PASSO' . $HOUSEmetnum}, $datareturn);
                $datareturn                = null;
                $previous['prevSELF']      = retrievecsv($SELFCmetnum, $lines, ${'PASSO' . $SELFCmetnum}, $datareturn);
                $datareturn                = null;
                $previous['prevPRODhouse'] = retrievecsv($PRODmetnum, $lines, ${'PASSO' . $PRODmetnum}, $datareturn);
                $previous['prevPRODself']  = $previous['prevPRODhouse'];
            } else { // restarting from scratch !
                $import     = null;
                $export     = null;
                $production = null;
                // latest import
                $import     = getvalue($IMPID, $IMPcmd);
                // latest export
                $export     = getvalue($EXPID, $EXPcmd);
                // latest production
                $production = getvalue($PRODID, $PRODcmd);
                
                $previous['prevIMPhouse']  = $import;
                $previous['prevEXPhouse']  = $export;
                $previous['prevEXPself']   = $export;
                $previous['prevHOUSE']     = 0;
                $previous['prevSELF']      = 0;
                $previous['prevPRODhouse'] = $production;
                $previous['prevPRODself']  = $production;
            }
        }
        
        // Now retrieve latest values
        $import     = null;
        $export     = null;
        $production = null;
        $outstr     = null;
        
        if ($argv[1] == 'energy') { // latest import
            $import = getvalue($IMPID, $IMPcmd);
        }
        $export     = getvalue($EXPID, $EXPcmd); // latest export
        $production = getvalue($PRODID, $PRODcmd); // latest production
        
        // Household consumption
        if ($argv[1] == 'energy' && isset($import) && isset($export)) {
            if ($export >= $previous['prevEXPhouse']) { // Some passover checks
                $diffEXP = $export - $previous['prevEXPhouse'];
            } else {
                $diffEXP = $export + ${'PASSO' . $EXPmetnum} - $previous['prevEXPhouse'];
            }
            if (isset($production)) {
                if ($production >= $previous['prevPRODhouse']) {
                    $diffPROD = $production - $previous['prevPRODhouse'];
                } else {
                    $diffPROD = $production + ${'PASSO' . $PRODmetnum} - $previous['prevPRODhouse'];
                }
                settype($previous['prevPRODhouse'], 'int');
                $previous['prevPRODhouse'] = $production;
            } else { // no production case
                $diffPROD = 0;
                $diffEXP  = 0;
            }
            if ($import >= $previous['prevIMPhouse']) {
                $diffIMP = $import - $previous['prevIMPhouse'];
            } else {
                $diffIMP = $import + ${'PASSO' . $IMPmetnum} - $previous['prevIMPhouse'];
            }
            $difference = $diffIMP + $diffPROD - $diffEXP;
            if ($difference < 0) { // Might happen if the production return no difference while the export did !
                $difference = 0;
            }
            $previous['prevHOUSE'] += $difference;
            
            if ($previous['prevHOUSE'] >= ${'PASSO' . $HOUSEmetnum}) { // passed over
                $previous['prevHOUSE'] -= ${'PASSO' . $HOUSEmetnum};
            }
            $val    = $previous['prevHOUSE'];
            $outstr = utf8_decode("$HOUSEID($val*Wh)\n");
            
            settype($previous['prevIMPhouse'], 'int');
            $previous['prevIMPhouse'] = $import;
            settype($previous['prevEXPhouse'], 'int');
            $previous['prevEXPhouse'] = $export;
            settype($previous['prevHOUSE'], 'int');
        } elseif ($argv[1] == 'self' && isset($export)) { // Self 
            // Some passover checks
            if ($export >= $previous['prevEXPself']) {
                $diffEXP = $export - $previous['prevEXPself'];
            } else {
                $diffEXP = $export + ${'PASSO' . $EXPmetnum} - $previous['prevEXPself'];
            }
            if (isset($production)) {
                if ($production >= $previous['prevPRODself']) {
                    $diffPROD = $production - $previous['prevPRODself'];
                } else {
                    $diffPROD = $production + ${'PASSO' . $PRODmetnum} - $previous['prevPRODself'];
                }
                settype($previous['prevPRODself'], 'int');
                $previous['prevPRODself'] = $production;
            } else { // no production case
                $diffPROD = 0;
                $diffEXP  = 0;
            }
            
            $difference = $diffPROD - $diffEXP;
            if ($difference < 0) {
                $difference = 0;
            }
            
            $previous['prevSELF'] += $difference;
            if ($previous['prevSELF'] >= ${'PASSO' . $SELFCmetnum}) {
                $previous['prevSELF'] -= ${'PASSO' . $SELFCmetnum};
            }
            $val    = $previous['prevSELF'];
            $outstr = utf8_decode("$SELFID($val*Wh)\n");
            
            settype($previous['prevEXPself'], 'int');
            $previous['prevEXPself'] = $export;
            settype($previous['prevSELF'], 'int');
        }
        
        // Saving previous values
        $data = json_encode($previous);
        file_put_contents($prevfile, $data);
        
        echo "$outstr";
    } else {
        die("Abording: no valid argument given\n");
    }
} else {
    echo "Usage: houseenergy { power | powerimp | powerexp | powerself | energy | self }\n";
    if (file_exists($prevfile)) {
        $data     = file_get_contents($prevfile);
        $previous = json_decode($data, true);
        print_r($previous);
    }
}
?>
